# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath(".."))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Congrads"
copyright = "%Y, DTAI KU Leuven"
author = "Wout Rombouts, Quinten Van Baelen, Peter Karsmakers"

try:
    from importlib.metadata import version as get_version  # Python 3.8+
except ImportError:
    from pkg_resources import (
        get_distribution as get_version,
    )  # Fallback for older versions

try:
    release = get_version("congrads")  # Replace with your package name
except Exception:
    release = "0.0.0"  # Fallback if the package isn't installed

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.duration",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "pytorch": ("https://pytorch.org/docs/stable", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "pandas": ("https://pandas.pydata.org/docs", None),
}
intersphinx_disabled_domains = ["std"]

templates_path = ["_templates"]

# -- Options for EPUB output
epub_show_urls = "footnote"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]

html_favicon = "_static/congrads_favicon.png"

html_theme_options = {
    "logo_only": False,
    "style_nav_header_background": "#cc5353",
}



"""
This module defines the `Descriptor` class, which is designed to manage
the mapping between neuron names, their corresponding layers, and additional
properties such as constant or variable status. It provides a way to easily 
place constraints on parts of your network, by referencing the neuron names
instead of indices.

The `Descriptor` class allows for easy constraint definitions on parts of
your neural network. It supports registering neurons with associated layers, 
indices, and optional attributes, such as whether the layer is constant 
or variable.

Key Methods:

    - `__init__`: Initializes the `Descriptor` object with empty mappings
      and sets for managing neurons and layers.
    - `add`: Registers a neuron with its associated layer, index, and
      optional constant status.

"""

from .utils import validate_type


class Descriptor:
    """
    A class to manage the mapping between neuron names, their corresponding
    layers, and additional properties (such as min/max values, output,
    and constant variables).

    This class is designed to track the relationship between neurons and
    layers in a neural network. It allows for the assignment of properties
    (like minimum and maximum values, and whether a layer is an output,
    constant, or variable) to each neuron. The data is stored in
    dictionaries and sets for efficient lookups.

    Attributes:
        neuron_to_layer (dict): A dictionary mapping neuron names to
            their corresponding layer names.
        neuron_to_index (dict): A dictionary mapping neuron names to
            their corresponding indices in the layers.
        constant_layers (set): A set of layer names that represent
            constant layers.
        variable_layers (set): A set of layer names that represent
            variable layers.
    """

    def __init__(
        self,
    ):
        """
        Initializes the Descriptor object.
        """

        # Define dictionaries that will translate neuron
        # names to layer and index
        self.neuron_to_layer: dict[str, str] = {}
        self.neuron_to_index: dict[str, int] = {}

        # Define sets that will hold the layers based on which type
        self.constant_layers: set[str] = set()
        self.variable_layers: set[str] = set()

    def add(
        self,
        layer_name: str,
        index: int,
        neuron_name: str,
        constant: bool = False,
    ):
        """
        Adds a neuron to the descriptor with its associated layer,
        index, and properties.

        This method registers a neuron name and associates it with a
        layer, its index, and optional properties such as whether
        the layer is an output or constant layer.

        Args:
            layer_name (str): The name of the layer where the neuron is located.
            index (int): The index of the neuron within the layer.
            neuron_name (str): The name of the neuron.
            constant (bool, optional): Whether the layer is a constant layer.
                Defaults to False.

        Raises:
            TypeError: If a provided attribute has an incompatible type.
            ValueError: If a layer or index is already assigned for a neuron
                or a duplicate index is used within a layer.

        """

        # Type checking
        validate_type("layer_name", layer_name, str)
        validate_type("index", index, int)
        validate_type("neuron_name", neuron_name, str)
        validate_type("constant", constant, bool)

        # Other validations
        if neuron_name in self.neuron_to_layer:
            raise ValueError(
                "There already is a layer registered for the neuron with name "
                f"'{neuron_name}'. Please use a unique name for each neuron."
            )

        if neuron_name in self.neuron_to_index:
            raise ValueError(
                "There already is an index registered for the neuron with name "
                f"'{neuron_name}'. Please use a unique name for each neuron."
            )

        for existing_neuron, assigned_index in self.neuron_to_index.items():
            if (
                assigned_index == index
                and self.neuron_to_layer[existing_neuron] == layer_name
            ):
                raise ValueError(
                    f"The index {index} in layer {layer_name} is already "
                    "assigned. Every neuron must be assigned a different "
                    "index that matches the network's output."
                )

        # Add to dictionaries and sets
        if constant:
            self.constant_layers.add(layer_name)
        else:
            self.variable_layers.add(layer_name)

        self.neuron_to_layer[neuron_name] = layer_name
        self.neuron_to_index[neuron_name] = index



"""
This module provides a set of constraint classes for guiding neural network 
training by enforcing specific conditions on the network's outputs.

The constraints in this module include:

- `Constraint`: The base class for all constraint types, defining the 
  interface and core behavior.
- `ImplicationConstraint`: A constraint that enforces one condition only if 
  another condition is met, useful for modeling implications between network 
  outputs.
- `ScalarConstraint`: A constraint that enforces scalar-based comparisons on 
  a network's output.
- `BinaryConstraint`: A constraint that enforces a binary comparison between
  two neurons in the network, using a comparison function (e.g., less than,
  greater than).
- `SumConstraint`: A constraint that enforces that the sum of certain neurons'
  outputs equals a specified value, which can be used to control total output.
- `PythagoreanConstraint`: A constraint that enforces the Pythagorean theorem
  on a set of neurons, ensuring that the square of one neuron's output is equal
  to the sum of the squares of other outputs.

These constraints can be used to steer the learning process by applying 
conditions such as logical implications or numerical bounds.

Usage:
    1. Define a custom constraint class by inheriting from `Constraint`.
    2. Apply the constraint to your neural network during training to 
       enforce desired output behaviors.
    3. Use the helper classes like `IdentityTransformation` for handling 
       transformations and comparisons in constraints.

Dependencies:
    - PyTorch (`torch`)
"""

import random
import string
import warnings
from abc import ABC, abstractmethod
from numbers import Number
from typing import Callable, Dict, Union

from torch import (
    Tensor,
    count_nonzero,
    ge,
    gt,
    isclose,
    le,
    logical_not,
    logical_or,
    lt,
    numel,
    ones,
    ones_like,
    reshape,
    sign,
    sqrt,
    square,
    stack,
    tensor,
    zeros_like,
)
from torch.nn.functional import normalize

from .descriptor import Descriptor
from .transformations import IdentityTransformation, Transformation
from .utils import validate_comparator_pytorch, validate_iterable, validate_type


class Constraint(ABC):
    """
    Abstract base class for defining constraints applied to neural networks.

    A `Constraint` specifies conditions that the neural network outputs
    should satisfy. It supports monitoring constraint satisfaction
    during training and can adjust loss to enforce constraints. Subclasses
    must implement the `check_constraint` and `calculate_direction` methods.

    Args:
        neurons (set[str]): Names of the neurons this constraint applies to.
        name (str, optional): A unique name for the constraint. If not provided,
            a name is generated based on the class name and a random suffix.
        monitor_only (bool, optional): If True, only monitor the constraint
            without adjusting the loss. Defaults to False.
        rescale_factor (Number, optional): Factor to scale the
            constraint-adjusted loss. Defaults to 1.5. Should be greater
            than 1 to give weight to the constraint.

    Raises:
        TypeError: If a provided attribute has an incompatible type.
        ValueError: If any neuron in `neurons` is not
            defined in the `descriptor`.

    Note:
        - If `rescale_factor <= 1`, a warning is issued, and the value is
          adjusted to a positive value greater than 1.
        - If `name` is not provided, a name is auto-generated,
          and a warning is logged.

    """

    descriptor: Descriptor = None
    device = None

    def __init__(
        self,
        neurons: set[str],
        name: str = None,
        monitor_only: bool = False,
        rescale_factor: Number = 1.5,
    ) -> None:
        """
        Initializes a new Constraint instance.
        """

        # Init parent class
        super().__init__()

        # Type checking
        validate_iterable("neurons", neurons, str)
        validate_type("name", name, (str, type(None)))
        validate_type("monitor_only", monitor_only, bool)
        validate_type("rescale_factor", rescale_factor, Number)

        # Init object variables
        self.neurons = neurons
        self.rescale_factor = rescale_factor
        self.monitor_only = monitor_only

        # Perform checks
        if rescale_factor <= 1:
            warnings.warn(
                "Rescale factor for constraint %s is <= 1. The network \
                    will favor general loss over the constraint-adjusted loss. \
                    Is this intended behavior? Normally, the loss should \
                    always be larger than 1.",
                name,
            )

        # If no constraint_name is set, generate one based
        # on the class name and a random suffix
        if name:
            self.name = name
        else:
            random_suffix = "".join(
                random.choices(string.ascii_uppercase + string.digits, k=6)
            )
            self.name = f"{self.__class__.__name__}_{random_suffix}"
            warnings.warn(
                "Name for constraint is not set. Using %s.", self.name
            )

        # If rescale factor is not larger than 1, warn user and adjust
        if rescale_factor <= 1:
            self.rescale_factor = abs(rescale_factor) + 1.5
            warnings.warn(
                "Rescale factor for constraint %s is < 1, adjusted value \
                    %s to %s.",
                name,
                rescale_factor,
                self.rescale_factor,
            )
        else:
            self.rescale_factor = rescale_factor

        # Infer layers from descriptor and neurons
        self.layers = set()
        for neuron in self.neurons:
            if neuron not in self.descriptor.neuron_to_layer.keys():
                raise ValueError(
                    f'The neuron name {neuron} used with constraint \
                        {self.name} is not defined in the descriptor. Please \
                        add it to the correct layer using \
                        descriptor.add("layer", ...).'
                )

            self.layers.add(self.descriptor.neuron_to_layer[neuron])

    @abstractmethod
    def check_constraint(
        self, prediction: dict[str, Tensor]
    ) -> tuple[Tensor, int]:
        """
        Evaluates whether the given model predictions satisfy the constraint.

        Args:
            prediction (dict[str, Tensor]): Model predictions for the neurons.

        Returns:
            tuple[Tensor, int]: A tuple where the first element is a tensor
            indicating whether the constraint is satisfied (with `True`
            for satisfaction, `False` for non-satisfaction, and `torch.nan`
            for irrelevant results), and the second element is an integer
            value representing the number of relevant constraints.

        Raises:
            NotImplementedError: If not implemented in a subclass.
        """

        raise NotImplementedError

    @abstractmethod
    def calculate_direction(
        self, prediction: dict[str, Tensor]
    ) -> Dict[str, Tensor]:
        """
        Calculates adjustment directions for neurons to
        better satisfy the constraint.

        Args:
            prediction (dict[str, Tensor]): Model predictions for the neurons.

        Returns:
            Dict[str, Tensor]: Dictionary mapping neuron layers to tensors
            specifying the adjustment direction for each neuron.

        Raises:
            NotImplementedError: If not implemented in a subclass.
        """

        raise NotImplementedError


class ImplicationConstraint(Constraint):
    """
    Represents an implication constraint between two
    constraints (head and body).

    The implication constraint ensures that the `body` constraint only applies
    when the `head` constraint is satisfied. If the `head` constraint is not
    satisfied, the `body` constraint does not apply.

    Args:
        head (Constraint): The head of the implication. If this constraint
            is satisfied, the body constraint must also be satisfied.
        body (Constraint): The body of the implication. This constraint
            is enforced only when the head constraint is satisfied.
        name (str, optional): A unique name for the constraint. If not
            provided, the name is generated in the format
            "{body.name} if {head.name}". Defaults to None.
        monitor_only (bool, optional): If True, the constraint is only
            monitored without adjusting the loss. Defaults to False.
        rescale_factor (Number, optional): The scaling factor for the
            constraint-adjusted loss. Defaults to 1.5.

    Raises:
        TypeError: If a provided attribute has an incompatible type.

    """

    def __init__(
        self,
        head: Constraint,
        body: Constraint,
        name=None,
        monitor_only=False,
        rescale_factor=1.5,
    ):
        """
        Initializes an ImplicationConstraint instance.
        """

        # Type checking
        validate_type("head", head, Constraint)
        validate_type("body", body, Constraint)

        # Compose constraint name
        name = f"{body.name} if {head.name}"

        # Init parent class
        super().__init__(
            head.neurons | body.neurons,
            name,
            monitor_only,
            rescale_factor,
        )

        self.head = head
        self.body = body

    def check_constraint(
        self, prediction: dict[str, Tensor]
    ) -> tuple[Tensor, int]:

        # Check satisfaction of head and body constraints
        head_satisfaction, _ = self.head.check_constraint(prediction)
        body_satisfaction, _ = self.body.check_constraint(prediction)

        # If head constraint is satisfied (returning 1),
        # the body constraint matters (and should return 0/1 based on body)
        # If head constraint is not satisfied (returning 0),
        # the body constraint does not apply (and should return 1)
        result = logical_or(
            logical_not(head_satisfaction), body_satisfaction
        ).float()

        return result, count_nonzero(head_satisfaction)

    def calculate_direction(
        self, prediction: dict[str, Tensor]
    ) -> Dict[str, Tensor]:
        # NOTE currently only works for dense layers
        # due to neuron to index translation

        # Use directions of constraint body as update vector
        return self.body.calculate_direction(prediction)


class ScalarConstraint(Constraint):
    """
    A constraint that enforces scalar-based comparisons on a specific neuron.

    This class ensures that the output of a specified neuron satisfies a scalar
    comparison operation (e.g., less than, greater than, etc.). It uses a
    comparator function to validate the condition and calculates adjustment
    directions accordingly.

    Args:
        operand (Union[str, Transformation]): Name of the neuron or a
            transformation to apply.
        comparator (Callable[[Tensor, Number], Tensor]): A comparison
            function (e.g., `torch.ge`, `torch.lt`).
        scalar (Number): The scalar value to compare against.
        name (str, optional): A unique name for the constraint. If not
            provided, a name is auto-generated in the format
            "<neuron_name> <comparator> <scalar>".
        monitor_only (bool, optional): If True, only monitor the constraint
            without adjusting the loss. Defaults to False.
        rescale_factor (Number, optional): Factor to scale the
            constraint-adjusted loss. Defaults to 1.5.

    Raises:
        TypeError: If a provided attribute has an incompatible type.

    Notes:
        - The `neuron_name` must be defined in the `descriptor` mapping.
        - The constraint name is composed using the neuron name,
          comparator, and scalar value.

    """

    def __init__(
        self,
        operand: Union[str, Transformation],
        comparator: Callable[[Tensor, Number], Tensor],
        scalar: Number,
        name: str = None,
        monitor_only: bool = False,
        rescale_factor: Number = 1.5,
    ) -> None:
        """
        Initializes a ScalarConstraint instance.
        """

        # Type checking
        validate_type("operand", operand, (str, Transformation))
        validate_comparator_pytorch("comparator", comparator)
        validate_comparator_pytorch("comparator", comparator)
        validate_type("scalar", scalar, Number)

        # If transformation is provided, get neuron name,
        # else use IdentityTransformation
        if isinstance(operand, Transformation):
            neuron_name = operand.neuron_name
            transformation = operand
        else:
            neuron_name = operand
            transformation = IdentityTransformation(neuron_name)

        # Compose constraint name
        name = f"{neuron_name} {comparator.__name__} {str(scalar)}"

        # Init parent class
        super().__init__({neuron_name}, name, monitor_only, rescale_factor)

        # Init variables
        self.comparator = comparator
        self.scalar = scalar
        self.transformation = transformation

        # Get layer name and feature index from neuron_name
        self.layer = self.descriptor.neuron_to_layer[neuron_name]
        self.index = self.descriptor.neuron_to_index[neuron_name]

        # Calculate directions based on constraint operator
        if self.comparator in [lt, le]:
            self.direction = -1
        elif self.comparator in [gt, ge]:
            self.direction = 1

    def check_constraint(
        self, prediction: dict[str, Tensor]
    ) -> tuple[Tensor, int]:

        # Select relevant columns
        selection = prediction[self.layer][:, self.index]

        # Apply transformation
        selection = self.transformation(selection)

        # Calculate current constraint result
        result = self.comparator(selection, self.scalar).float()
        return result, numel(result)

    def calculate_direction(
        self, prediction: dict[str, Tensor]
    ) -> Dict[str, Tensor]:
        # NOTE currently only works for dense layers due
        # to neuron to index translation

        output = {}

        for layer in self.layers:
            output[layer] = zeros_like(prediction[layer][0], device=self.device)

        output[self.layer][self.index] = self.direction

        for layer in self.layers:
            output[layer] = normalize(reshape(output[layer], [1, -1]), dim=1)

        return output


class BinaryConstraint(Constraint):
    """
    A constraint that enforces a binary comparison between two neurons.

    This class ensures that the output of one neuron satisfies a comparison
    operation with the output of another neuron
    (e.g., less than, greater than, etc.). It uses a comparator function to
    validate the condition and calculates adjustment directions accordingly.

    Args:
        operand_left (Union[str, Transformation]): Name of the left
            neuron or a transformation to apply.
        comparator (Callable[[Tensor, Number], Tensor]): A comparison
            function (e.g., `torch.ge`, `torch.lt`).
        operand_right (Union[str, Transformation]): Name of the right
            neuron or a transformation to apply.
        name (str, optional): A unique name for the constraint. If not
            provided, a name is auto-generated in the format
            "<neuron_name_left> <comparator> <neuron_name_right>".
        monitor_only (bool, optional): If True, only monitor the constraint
            without adjusting the loss. Defaults to False.
        rescale_factor (Number, optional): Factor to scale the
            constraint-adjusted loss. Defaults to 1.5.

    Raises:
        TypeError: If a provided attribute has an incompatible type.

    Notes:
        - The neuron names must be defined in the `descriptor` mapping.
        - The constraint name is composed using the left neuron name,
          comparator, and right neuron name.

    """

    def __init__(
        self,
        operand_left: Union[str, Transformation],
        comparator: Callable[[Tensor, Number], Tensor],
        operand_right: Union[str, Transformation],
        name: str = None,
        monitor_only: bool = False,
        rescale_factor: Number = 1.5,
    ) -> None:
        """
        Initializes a BinaryConstraint instance.
        """

        # Type checking
        validate_type("operand_left", operand_left, (str, Transformation))
        validate_comparator_pytorch("comparator", comparator)
        validate_comparator_pytorch("comparator", comparator)
        validate_type("operand_right", operand_right, (str, Transformation))

        # If transformation is provided, get neuron name,
        # else use IdentityTransformation
        if isinstance(operand_left, Transformation):
            neuron_name_left = operand_left.neuron_name
            transformation_left = operand_left
        else:
            neuron_name_left = operand_left
            transformation_left = IdentityTransformation(neuron_name_left)

        if isinstance(operand_right, Transformation):
            neuron_name_right = operand_right.neuron_name
            transformation_right = operand_right
        else:
            neuron_name_right = operand_right
            transformation_right = IdentityTransformation(neuron_name_right)

        # Compose constraint name
        name = f"{neuron_name_left} {comparator.__name__} {neuron_name_right}"

        # Init parent class
        super().__init__(
            {neuron_name_left, neuron_name_right},
            name,
            monitor_only,
            rescale_factor,
        )

        # Init variables
        self.comparator = comparator
        self.transformation_left = transformation_left
        self.transformation_right = transformation_right

        # Get layer name and feature index from neuron_name
        self.layer_left = self.descriptor.neuron_to_layer[neuron_name_left]
        self.layer_right = self.descriptor.neuron_to_layer[neuron_name_right]
        self.index_left = self.descriptor.neuron_to_index[neuron_name_left]
        self.index_right = self.descriptor.neuron_to_index[neuron_name_right]

        # Calculate directions based on constraint operator
        if self.comparator in [lt, le]:
            self.direction_left = -1
            self.direction_right = 1
        else:
            self.direction_left = 1
            self.direction_right = -1

    def check_constraint(
        self, prediction: dict[str, Tensor]
    ) -> tuple[Tensor, int]:

        # Select relevant columns
        selection_left = prediction[self.layer_left][:, self.index_left]
        selection_right = prediction[self.layer_right][:, self.index_right]

        # Apply transformations
        selection_left = self.transformation_left(selection_left)
        selection_right = self.transformation_right(selection_right)

        result = self.comparator(selection_left, selection_right).float()

        return result, numel(result)

    def calculate_direction(
        self, prediction: dict[str, Tensor]
    ) -> Dict[str, Tensor]:
        # NOTE currently only works for dense layers due
        # to neuron to index translation

        output = {}

        for layer in self.layers:
            output[layer] = zeros_like(prediction[layer][0], device=self.device)

        output[self.layer_left][self.index_left] = self.direction_left
        output[self.layer_right][self.index_right] = self.direction_right

        for layer in self.layers:
            output[layer] = normalize(reshape(output[layer], [1, -1]), dim=1)

        return output


class SumConstraint(Constraint):
    """
    A constraint that enforces a weighted summation comparison
    between two groups of neurons.

    This class evaluates whether the weighted sum of outputs from one set of
    neurons satisfies a comparison operation with the weighted sum of
    outputs from another set of neurons.

    Args:
        operands_left (list[Union[str, Transformation]]): List of neuron
            names or transformations on the left side.
        comparator (Callable[[Tensor, Number], Tensor]): A comparison
            function for the constraint.
        operands_right (list[Union[str, Transformation]]): List of neuron
            names or transformations on the right side.
        weights_left (list[Number], optional): Weights for the left neurons.
            Defaults to None.
        weights_right (list[Number], optional): Weights for the right
            neurons. Defaults to None.
        name (str, optional): Unique name for the constraint.
            If None, it's auto-generated. Defaults to None.
        monitor_only (bool, optional): If True, only monitor the constraint
            without adjusting the loss. Defaults to False.
        rescale_factor (Number, optional): Factor to scale the
            constraint-adjusted loss. Defaults to 1.5.

    Raises:
        TypeError: If a provided attribute has an incompatible type.
        ValueError: If the dimensions of neuron names and weights mismatch.

    """

    def __init__(
        self,
        operands_left: list[Union[str, Transformation]],
        comparator: Callable[[Tensor, Number], Tensor],
        operands_right: list[Union[str, Transformation]],
        weights_left: list[Number] = None,
        weights_right: list[Number] = None,
        name: str = None,
        monitor_only: bool = False,
        rescale_factor: Number = 1.5,
    ) -> None:
        """
        Initializes the SumConstraint.
        """

        # Type checking
        validate_iterable("operands_left", operands_left, (str, Transformation))
        validate_comparator_pytorch("comparator", comparator)
        validate_comparator_pytorch("comparator", comparator)
        validate_iterable(
            "operands_right", operands_right, (str, Transformation)
        )
        validate_iterable("weights_left", weights_left, Number, allow_none=True)
        validate_iterable(
            "weights_right", weights_right, Number, allow_none=True
        )

        # If transformation is provided, get neuron name,
        # else use IdentityTransformation
        neuron_names_left: list[str] = []
        transformations_left: list[Transformation] = []
        for operand_left in operands_left:
            if isinstance(operand_left, Transformation):
                neuron_name_left = operand_left.neuron_name
                neuron_names_left.append(neuron_name_left)
                transformations_left.append(operand_left)
            else:
                neuron_name_left = operand_left
                neuron_names_left.append(neuron_name_left)
                transformations_left.append(
                    IdentityTransformation(neuron_name_left)
                )

        neuron_names_right: list[str] = []
        transformations_right: list[Transformation] = []
        for operand_right in operands_right:
            if isinstance(operand_right, Transformation):
                neuron_name_right = operand_right.neuron_name
                neuron_names_right.append(neuron_name_right)
                transformations_right.append(operand_right)
            else:
                neuron_name_right = operand_right
                neuron_names_right.append(neuron_name_right)
                transformations_right.append(
                    IdentityTransformation(neuron_name_right)
                )

        # Compose constraint name
        w_left = weights_left or [""] * len(neuron_names_left)
        w_right = weights_right or [""] * len(neuron_names_right)
        left_expr = " + ".join(
            f"{w}{n}" for w, n in zip(w_left, neuron_names_left)
        )
        right_expr = " + ".join(
            f"{w}{n}" for w, n in zip(w_right, neuron_names_right)
        )
        comparator_name = comparator.__name__
        name = f"{left_expr} {comparator_name} {right_expr}"

        # Init parent class
        neuron_names = set(neuron_names_left) | set(neuron_names_right)
        super().__init__(neuron_names, name, monitor_only, rescale_factor)

        # Init variables
        self.comparator = comparator
        self.neuron_names_left = neuron_names_left
        self.neuron_names_right = neuron_names_right
        self.transformations_left = transformations_left
        self.transformations_right = transformations_right

        # If feature list dimensions don't match
        # weight list dimensions, raise error
        if weights_left and (len(neuron_names_left) != len(weights_left)):
            raise ValueError(
                "The dimensions of neuron_names_left don't match with the \
                    dimensions of weights_left."
            )
        if weights_right and (len(neuron_names_right) != len(weights_right)):
            raise ValueError(
                "The dimensions of neuron_names_right don't match with the \
                    dimensions of weights_right."
            )

        # If weights are provided for summation, transform them to Tensors
        if weights_left:
            self.weights_left = tensor(weights_left, device=self.device)
        else:
            self.weights_left = ones(len(neuron_names_left), device=self.device)
        if weights_right:
            self.weights_right = tensor(weights_right, device=self.device)
        else:
            self.weights_right = ones(
                len(neuron_names_right), device=self.device
            )

        # Calculate directions based on constraint operator
        if self.comparator in [lt, le]:
            self.direction_left = -1
            self.direction_right = 1
        else:
            self.direction_left = 1
            self.direction_right = -1

    def check_constraint(
        self, prediction: dict[str, Tensor]
    ) -> tuple[Tensor, int]:

        def compute_weighted_sum(
            neuron_names: list[str],
            transformations: list[Transformation],
            weights: tensor,
        ) -> tensor:
            layers = [
                self.descriptor.neuron_to_layer[neuron_name]
                for neuron_name in neuron_names
            ]
            indices = [
                self.descriptor.neuron_to_index[neuron_name]
                for neuron_name in neuron_names
            ]

            # Select relevant column
            selections = [
                prediction[layer][:, index]
                for layer, index in zip(layers, indices)
            ]

            # Apply transformations
            results = []
            for transformation, selection in zip(transformations, selections):
                results.append(transformation(selection))

            # Extract predictions for all neurons and apply weights in bulk
            predictions = stack(
                results,
                dim=1,
            )

            # Calculate weighted sum
            return (predictions * weights.unsqueeze(0)).sum(dim=1)

        # Compute weighted sums
        weighted_sum_left = compute_weighted_sum(
            self.neuron_names_left,
            self.transformations_left,
            self.weights_left,
        )
        weighted_sum_right = compute_weighted_sum(
            self.neuron_names_right,
            self.transformations_right,
            self.weights_right,
        )

        # Apply the comparator and calculate the result
        result = self.comparator(weighted_sum_left, weighted_sum_right).float()

        return result, numel(result)

    def calculate_direction(
        self, prediction: dict[str, Tensor]
    ) -> Dict[str, Tensor]:
        # NOTE currently only works for dense layers
        # due to neuron to index translation

        output = {}

        for layer in self.layers:
            output[layer] = zeros_like(prediction[layer][0], device=self.device)

        for neuron_name_left in self.neuron_names_left:
            layer = self.descriptor.neuron_to_layer[neuron_name_left]
            index = self.descriptor.neuron_to_index[neuron_name_left]
            output[layer][index] = self.direction_left

        for neuron_name_right in self.neuron_names_right:
            layer = self.descriptor.neuron_to_layer[neuron_name_right]
            index = self.descriptor.neuron_to_index[neuron_name_right]
            output[layer][index] = self.direction_right

        for layer in self.layers:
            output[layer] = normalize(reshape(output[layer], [1, -1]), dim=1)

        return output


class PythagoreanIdentityConstraint(Constraint):
    """
    A constraint that enforces the Pythagorean identity: a² + b² ≈ 1,
    where `a` and `b` are neurons or transformations.

    This constraint checks that the sum of the squares of two specified
    neurons (or their transformations) is approximately equal to 1.
    The constraint is evaluated using relative and absolute
    tolerance (`rtol` and `atol`) and is applied during the forward pass.

    Args:
        a (Union[str, Transformation]): The first input, either a
            neuron name (str) or a Transformation.
        b (Union[str, Transformation]): The second input, either a
            neuron name (str) or a Transformation.
        rtol (float, optional): The relative tolerance for the
            comparison (default is 0.00001).
        atol (float, optional): The absolute tolerance for the
            comparison (default is 1e-8).
        name (str, optional): The name of the constraint
            (default is None, and it is generated automatically).
        monitor_only (bool, optional): Flag indicating whether the
            constraint is only for monitoring (default is False).
        rescale_factor (Number, optional): A factor used for
            rescaling (default is 1.5).

    Raises:
        TypeError: If a provided attribute has an incompatible type.

    """

    def __init__(
        self,
        a: Union[str, Transformation],
        b: Union[str, Transformation],
        rtol: float = 0.00001,
        atol: float = 1e-8,
        name: str = None,
        monitor_only: bool = False,
        rescale_factor: Number = 1.5,
    ) -> None:
        """
        Initialize the PythagoreanIdentityConstraint.
        """

        # Type checking
        validate_type("a", a, (str, Transformation))
        validate_type("b", b, (str, Transformation))
        validate_type("rtol", rtol, float)
        validate_type("atol", atol, float)

        # If transformation is provided, get neuron name,
        # else use IdentityTransformation
        if isinstance(a, Transformation):
            neuron_name_a = a.neuron_name
            transformation_a = a
        else:
            neuron_name_a = a
            transformation_a = IdentityTransformation(neuron_name_a)

        if isinstance(b, Transformation):
            neuron_name_b = b.neuron_name
            transformation_b = b
        else:
            neuron_name_b = b
            transformation_b = IdentityTransformation(neuron_name_b)

        # Compose constraint name
        name = f"{neuron_name_a}² + {neuron_name_b}² ≈ 1"

        # Init parent class
        super().__init__(
            {neuron_name_a, neuron_name_b},
            name,
            monitor_only,
            rescale_factor,
        )

        # Init variables
        self.transformation_a = transformation_a
        self.transformation_b = transformation_b
        self.rtol = rtol
        self.atol = atol

        # Get layer name and feature index from neuron_name
        self.layer_a = self.descriptor.neuron_to_layer[neuron_name_a]
        self.layer_b = self.descriptor.neuron_to_layer[neuron_name_b]
        self.index_a = self.descriptor.neuron_to_index[neuron_name_a]
        self.index_b = self.descriptor.neuron_to_index[neuron_name_b]

    def check_constraint(
        self, prediction: dict[str, Tensor]
    ) -> tuple[Tensor, int]:

        # Select relevant columns
        selection_a = prediction[self.layer_a][:, self.index_a]
        selection_b = prediction[self.layer_b][:, self.index_b]

        # Apply transformations
        selection_a = self.transformation_a(selection_a)
        selection_b = self.transformation_b(selection_b)

        # Calculate result
        result = isclose(
            square(selection_a) + square(selection_b),
            ones_like(selection_a, device=self.device),
            rtol=self.rtol,
            atol=self.atol,
        ).float()

        return result, numel(result)

    def calculate_direction(
        self, prediction: dict[str, Tensor]
    ) -> Dict[str, Tensor]:
        # NOTE currently only works for dense layers due
        # to neuron to index translation

        output = {}

        for layer in self.layers:
            output[layer] = zeros_like(prediction[layer], device=self.device)

        a = prediction[self.layer_a][:, self.index_a]
        b = prediction[self.layer_b][:, self.index_b]
        m = sqrt(square(a) + square(b))

        output[self.layer_a][:, self.index_a] = a / m * sign(1 - m)
        output[self.layer_b][:, self.index_b] = b / m * sign(1 - m)

        return output



"""
This module provides utility functions and classes for managing data,
logging, and preprocessing in machine learning workflows. The functionalities
include logging key-value pairs to CSV files, splitting datasets into
training, validation, and test sets, preprocessing functions for various 
data sets and validator functions for type checking.
"""

import os

import numpy as np
import pandas as pd
import torch
from torch import Generator
from torch.utils.data import DataLoader, Dataset, random_split


class CSVLogger:
    """
    A utility class for logging key-value pairs to a CSV file, organized by
    epochs. Supports merging with existing logs or overwriting them.

    Args:
        file_path (str): The path to the CSV file for logging.
        overwrite (bool): If True, overwrites any existing file at
            the file_path.
        merge (bool): If True, merges new values with existing data
            in the file.

    Raises:
        ValueError: If both overwrite and merge are True.
        FileExistsError: If the file already exists and neither
            overwrite nor merge is True.
    """

    def __init__(
        self, file_path: str, overwrite: bool = False, merge: bool = True
    ):
        """
        Initializes the CSVLogger.
        """

        self.file_path = file_path
        self.values: dict[tuple[int, str], float] = {}

        if merge and overwrite:
            raise ValueError(
                "The attributes overwrite and merge cannot be True at the "
                "same time. Either specify overwrite=True or merge=True."
            )

        if not os.path.exists(file_path):
            pass
        elif merge:
            self.load()
        elif overwrite:
            pass
        else:
            raise FileExistsError(
                f"A CSV file already exists at {file_path}. Specify "
                "CSVLogger(..., overwrite=True) to overwrite the file."
            )

    def add_value(self, name: str, value: float, epoch: int):
        """
        Adds a value to the logger for a specific epoch and name.

        Args:
            name (str): The name of the metric or value to log.
            value (float): The value to log.
            epoch (int): The epoch associated with the value.
        """

        self.values[epoch, name] = value

    def save(self):
        """
        Saves the logged values to the specified CSV file.

        If the file exists and merge is enabled, merges the current data
        with the existing file.
        """

        data = self.to_dataframe(self.values)
        data.to_csv(self.file_path, index=False)

    def load(self):
        """
        Loads data from the CSV file into the logger.

        Converts the CSV data into the internal dictionary format for
        further updates or operations.
        """

        df = pd.read_csv(self.file_path)
        self.values = self.to_dict(df)

    @staticmethod
    def to_dataframe(values: dict[tuple[int, str], float]) -> pd.DataFrame:
        """
        Converts a dictionary of values into a DataFrame.

        Args:
            values (dict[tuple[int, str], float]): A dictionary of values
                keyed by (epoch, name).

        Returns:
            pd.DataFrame: A DataFrame where epochs are rows, names are
            columns, and values are the cell data.
        """

        # Convert to a DataFrame
        df = pd.DataFrame.from_dict(values, orient="index", columns=["value"])

        # Reset the index to separate epoch and name into columns
        df.index = pd.MultiIndex.from_tuples(df.index, names=["epoch", "name"])
        df = df.reset_index()

        # Pivot the DataFrame so epochs are rows and names are columns
        result = df.pivot(index="epoch", columns="name", values="value")

        # Optional: Reset the column names for a cleaner look
        result = result.reset_index().rename_axis(columns=None)

        return result

    @staticmethod
    def to_dict(df: pd.DataFrame) -> dict[tuple[int, str], float]:
        """
        Converts a DataFrame with epochs as rows and names as columns
        back into a dictionary of the format {(epoch, name): value}.
        """
        # Set the epoch column as the index (if not already)
        df = df.set_index("epoch")

        # Stack the DataFrame to create a multi-index series
        stacked = df.stack()

        # Convert the multi-index series to a dictionary
        result = stacked.to_dict()

        return result


def split_data_loaders(
    data: Dataset,
    loader_args: dict = None,
    train_loader_args: dict = None,
    valid_loader_args: dict = None,
    test_loader_args: dict = None,
    train_size: float = 0.8,
    valid_size: float = 0.1,
    test_size: float = 0.1,
    split_generator: Generator = None,
) -> tuple[DataLoader, DataLoader, DataLoader]:
    """
    Splits a dataset into training, validation, and test sets,
    and returns corresponding DataLoader objects.

    Args:
        data (Dataset): The dataset to be split.
        loader_args (dict, optional): Default DataLoader arguments, merges
            with loader-specific arguments, overlapping keys from
            loader-specific arguments are superseded.
        train_loader_args (dict, optional): Training DataLoader arguments,
            merges with `loader_args`, overriding overlapping keys.
        valid_loader_args (dict, optional): Validation DataLoader arguments,
            merges with `loader_args`, overriding overlapping keys.
        test_loader_args (dict, optional): Test DataLoader arguments,
            merges with `loader_args`, overriding overlapping keys.
        train_size (float, optional): Proportion of data to be used for
            training. Defaults to 0.8.
        valid_size (float, optional): Proportion of data to be used for
            validation. Defaults to 0.1.
        test_size (float, optional): Proportion of data to be used for
            testing. Defaults to 0.1.
        split_generator (Generator, optional): Optional random seed generator
            to control the splitting of the dataset.

    Returns:
        tuple: A tuple containing three DataLoader objects: one for the
        training, validation and test set.

    Raises:
        ValueError: If the train_size, valid_size, and test_size are not
            between 0 and 1, or if their sum does not equal 1.
    """

    # Validate split sizes
    if not (0 < train_size < 1 and 0 < valid_size < 1 and 0 < test_size < 1):
        raise ValueError(
            "train_size, valid_size, and test_size must be between 0 and 1."
        )
    if not abs(train_size + valid_size + test_size - 1.0) < 1e-6:
        raise ValueError("train_size, valid_size, and test_size must sum to 1.")

    # Perform the splits
    train_val_data, test_data = random_split(
        data, [1 - test_size, test_size], generator=split_generator
    )
    train_data, valid_data = random_split(
        train_val_data,
        [
            train_size / (1 - test_size),
            valid_size / (1 - test_size),
        ],
        generator=split_generator,
    )

    # Set default arguments for each loader
    train_loader_args = dict(loader_args or {}, **(train_loader_args or {}))
    valid_loader_args = dict(loader_args or {}, **(valid_loader_args or {}))
    test_loader_args = dict(loader_args or {}, **(test_loader_args or {}))

    # Create the DataLoaders
    train_generator = DataLoader(train_data, **train_loader_args)
    valid_generator = DataLoader(valid_data, **valid_loader_args)
    test_generator = DataLoader(test_data, **test_loader_args)

    return train_generator, valid_generator, test_generator


# pylint: disable-next=invalid-name
def preprocess_BiasCorrection(df: pd.DataFrame) -> pd.DataFrame:
    """
    Preprocesses the given dataframe for bias correction by
    performing a series of transformations.

    The function sequentially:

    - Drops rows with missing values.
    - Converts a date string to datetime format and adds year, month,
      and day columns.
    - Normalizes the columns with specific logic for input and output variables.
    - Adds a multi-index indicating which columns are input or output variables.
    - Samples 2500 examples from the dataset without replacement.

    Args:
        df (pd.DataFrame): The input dataframe containing the data
            to be processed.

    Returns:
        pd.DataFrame: The processed dataframe after applying
        the transformations.
    """

    def date_to_datetime(df: pd.DataFrame) -> pd.DataFrame:
        """Transform the string that denotes the date to
        the datetime format in pandas."""
        # make copy of dataframe
        df_temp = df.copy()
        # add new column at the front where the date string is
        # transformed to the datetime format
        df_temp.insert(0, "DateTransformed", pd.to_datetime(df_temp["Date"]))
        return df_temp

    def add_year(df: pd.DataFrame) -> pd.DataFrame:
        """Extract the year from the datetime cell and add it
        as a new column to the dataframe at the front."""
        # make copy of dataframe
        df_temp = df.copy()
        # extract year and add new column at the front containing these numbers
        df_temp.insert(0, "Year", df_temp["DateTransformed"].dt.year)
        return df_temp

    def add_month(df: pd.DataFrame) -> pd.DataFrame:
        """Extract the month from the datetime cell and add it
        as a new column to the dataframe at the front."""
        # make copy of dataframe
        df_temp = df.copy()
        # extract month and add new column at index 1 containing these numbers
        df_temp.insert(1, "Month", df_temp["DateTransformed"].dt.month)
        return df_temp

    def add_day(df: pd.DataFrame) -> pd.DataFrame:
        """Extract the day from the datetime cell and add it
        as a new column to the dataframe at the front."""
        # make copy of dataframe
        df_temp = df.copy()
        # extract day and add new column at index 2 containing these numbers
        df_temp.insert(2, "Day", df_temp["DateTransformed"].dt.day)
        return df_temp

    def add_input_output_temperature(df: pd.DataFrame) -> pd.DataFrame:
        """Add a multiindex denoting if the column
        is an input or output variable."""
        # copy the dataframe
        temp_df = df.copy()
        # extract all the column names
        column_names = temp_df.columns.tolist()
        # only the last 2 columns are output variables, all others are input
        # variables. So make list of corresponding lengths of
        # 'Input' and 'Output'
        input_list = ["Input"] * (len(column_names) - 2)
        output_list = ["Output"] * 2
        # concat both lists
        input_output_list = input_list + output_list
        # define multi index for attaching this 'Input' and 'Output' list with
        # the column names already existing
        multiindex_bias = pd.MultiIndex.from_arrays(
            [input_output_list, column_names]
        )
        # transpose such that index can be adjusted to multi index
        new_df = pd.DataFrame(df.transpose().to_numpy(), index=multiindex_bias)
        # transpose back such that columns are the same as before
        # except with different labels
        return new_df.transpose()

    def normalize_columns_bias(df: pd.DataFrame) -> pd.DataFrame:
        """Normalize the columns for the bias correction dataset.
        This is different from normalizing all the columns separately
        because the upper and lower bounds for the output variables
        are assumed to be the same."""
        # copy the dataframe
        temp_df = df.copy()
        # normalize each column
        for feature_name in df.columns:
            # the output columns are normalized using the same upper and
            # lower bound for more efficient check of the inequality
            if feature_name == "Next_Tmax" or feature_name == "Next_Tmin":
                max_value = 38.9
                min_value = 11.3
            # the input columns are normalized using their respective
            # upper and lower bounds
            else:
                max_value = df[feature_name].max()
                min_value = df[feature_name].min()
            temp_df[feature_name] = (df[feature_name] - min_value) / (
                max_value - min_value
            )
        return temp_df

    def sample_2500_examples(df: pd.DataFrame) -> pd.DataFrame:
        """Sample 2500 examples from the dataframe without replacement."""
        temp_df = df.copy()
        sample_df = temp_df.sample(
            n=2500, replace=False, random_state=3, axis=0
        )
        return sample_df

    return (
        # drop missing values
        df.dropna(how="any")
        # transform string date to datetime format
        .pipe(date_to_datetime)
        # add year as a single column
        .pipe(add_year)
        # add month as a single column
        .pipe(add_month)
        # add day as a single column
        .pipe(add_day)
        # remove original date string and the datetime format
        .drop(["Date", "DateTransformed"], axis=1, inplace=False)
        # convert all numbers to float32
        .astype("float32")
        # normalize columns
        .pipe(normalize_columns_bias)
        # add multi index indicating which columns are corresponding
        # to input and output variables
        .pipe(add_input_output_temperature)
        # sample 2500 examples out of the dataset
        .pipe(sample_2500_examples)
    )


# pylint: disable-next=invalid-name
def preprocess_FamilyIncome(df: pd.DataFrame) -> pd.DataFrame:
    """
    Preprocesses the given Family Income dataframe by applying a
    series of transformations and constraints.

    The function sequentially:

    - Drops rows with missing values.
    - Converts object columns to appropriate data types and
      removes string columns.
    - Removes certain unnecessary columns like
      'Agricultural Household indicator' and related features.
    - Adds labels to columns indicating whether they are
      input or output variables.
    - Normalizes the columns individually.
    - Checks and removes rows that do not satisfy predefined constraints
      (household income > expenses, food expenses > sub-expenses).
    - Samples 2500 examples from the dataset without replacement.

    Args:
        df (pd.DataFrame): The input Family Income dataframe containing
            the data to be processed.

    Returns:
        pd.DataFrame: The processed dataframe after applying the
        transformations and constraints.
    """

    def normalize_columns_income(df: pd.DataFrame) -> pd.DataFrame:
        """Normalize the columns for the Family Income dataframe.
        This can also be applied to other dataframes because this
        function normalizes all columns individually."""
        # copy the dataframe
        temp_df = df.copy()
        # normalize each column
        for feature_name in df.columns:
            max_value = df[feature_name].max()
            min_value = df[feature_name].min()
            temp_df[feature_name] = (df[feature_name] - min_value) / (
                max_value - min_value
            )
        return temp_df

    def check_constraints_income(df: pd.DataFrame) -> pd.DataFrame:
        """Check if all the constraints are satisfied for the dataframe
        and remove the examples that do not satisfy the constraint.
        This function only works for the Family Income dataset and the
        constraints are that the household income is larger than all the
        expenses and the food expense is larger than the sum of
        the other (more detailed) food expenses.
        """
        temp_df = df.copy()
        # check that household income is larger than expenses in the output
        input_array = temp_df["Input"].to_numpy()
        income_array = np.add(
            np.multiply(
                input_array[:, [0, 1]],
                np.subtract(
                    np.asarray([11815988, 9234485]), np.asarray([11285, 0])
                ),
            ),
            np.asarray([11285, 0]),
        )
        expense_array = temp_df["Output"].to_numpy()
        expense_array = np.add(
            np.multiply(
                expense_array,
                np.subtract(
                    np.asarray(
                        [
                            791848,
                            437467,
                            140992,
                            74800,
                            2188560,
                            1049275,
                            149940,
                            731000,
                        ]
                    ),
                    np.asarray([3704, 0, 0, 0, 1950, 0, 0, 0]),
                ),
            ),
            np.asarray([3704, 0, 0, 0, 1950, 0, 0, 0]),
        )
        expense_array_without_dup = expense_array[:, [0, 4, 5, 6, 7]]
        sum_expenses = np.sum(expense_array_without_dup, axis=1)
        total_income = np.sum(income_array, axis=1)
        sanity_check_array = np.greater_equal(total_income, sum_expenses)
        temp_df["Unimportant"] = sanity_check_array.tolist()
        reduction = temp_df[temp_df.Unimportant]
        drop_reduction = reduction.drop("Unimportant", axis=1)

        # check that the food expense is larger than all the sub expenses
        expense_reduced_array = drop_reduction["Output"].to_numpy()
        expense_reduced_array = np.add(
            np.multiply(
                expense_reduced_array,
                np.subtract(
                    np.asarray(
                        [
                            791848,
                            437467,
                            140992,
                            74800,
                            2188560,
                            1049275,
                            149940,
                            731000,
                        ]
                    ),
                    np.asarray([3704, 0, 0, 0, 1950, 0, 0, 0]),
                ),
            ),
            np.asarray([3704, 0, 0, 0, 1950, 0, 0, 0]),
        )
        food_mul_expense_array = expense_reduced_array[:, [1, 2, 3]]
        food_mul_expense_array_sum = np.sum(food_mul_expense_array, axis=1)
        food_expense_array = expense_reduced_array[:, 0]
        sanity_check_array = np.greater_equal(
            food_expense_array, food_mul_expense_array_sum
        )
        drop_reduction["Unimportant"] = sanity_check_array.tolist()
        new_reduction = drop_reduction[drop_reduction.Unimportant]
        satisfied_constraints_df = new_reduction.drop("Unimportant", axis=1)

        return satisfied_constraints_df

    def add_input_output_family_income(df: pd.DataFrame) -> pd.DataFrame:
        """Add a multiindex denoting if the column is
        an input or output variable."""

        # copy the dataframe
        temp_df = df.copy()
        # extract all the column names
        column_names = temp_df.columns.tolist()
        # the 2nd-9th columns correspond to output variables and all
        # others to input variables. So make list of corresponding
        # lengths of 'Input' and 'Output'
        input_list_start = ["Input"]
        input_list_end = ["Input"] * (len(column_names) - 9)
        output_list = ["Output"] * 8
        # concat both lists
        input_output_list = input_list_start + output_list + input_list_end
        # define multi index for attaching this 'Input' and
        # 'Output' list with the column names already existing
        multiindex_bias = pd.MultiIndex.from_arrays(
            [input_output_list, column_names]
        )
        # transpose such that index can be adjusted to multi index
        new_df = pd.DataFrame(df.transpose().to_numpy(), index=multiindex_bias)
        # transpose back such that columns are the same as
        # before except with different labels
        return new_df.transpose()

    def sample_2500_examples(df: pd.DataFrame) -> pd.DataFrame:
        """Sample 2500 examples from the dataframe without replacement."""
        temp_df = df.copy()
        sample_df = temp_df.sample(
            n=2500, replace=False, random_state=3, axis=0
        )
        return sample_df

    return (
        # drop missing values
        df.dropna(how="any")
        # convert object to fitting dtype
        .convert_dtypes()
        # remove all strings (no other dtypes are present
        # except for integers and floats)
        .select_dtypes(exclude=["string"])
        # transform all numbers to same dtype
        .astype("float32")
        # drop column with label Agricultural Household indicator
        # because this is not really a numerical input but
        # rather a categorical/classification
        .drop(["Agricultural Household indicator"], axis=1, inplace=False)
        # this column is dropped because it depends on
        # Agricultural Household indicator
        .drop(["Crop Farming and Gardening expenses"], axis=1, inplace=False)
        # use 8 output variables and 24 input variables
        .drop(
            [
                "Total Rice Expenditure",
                "Total Fish and  marine products Expenditure",
                "Fruit Expenditure",
                "Restaurant and hotels Expenditure",
                "Alcoholic Beverages Expenditure",
                "Tobacco Expenditure",
                "Clothing, Footwear and Other Wear Expenditure",
                "Imputed House Rental Value",
                "Transportation Expenditure",
                "Miscellaneous Goods and Services Expenditure",
                "Special Occasions Expenditure",
            ],
            axis=1,
            inplace=False,
        )
        # add input and output labels to each column
        .pipe(add_input_output_family_income)
        # normalize all the columns
        .pipe(normalize_columns_income)
        # remove all datapoints that do not satisfy the constraints
        .pipe(check_constraints_income)
        # sample 2500 examples
        .pipe(sample_2500_examples)
    )


def validate_type(name, value, expected_types, allow_none=False):
    """
    Validate that a value is of the specified type(s).

    Args:
        name (str): Name of the argument for error messages.
        value: Value to validate.
        expected_types (type or tuple of types): Expected type(s) for the value.
        allow_none (bool): Whether to allow the value to be None.
            Defaults to False.

    Raises:
        TypeError: If the value is not of the expected type(s).
    """

    if value is None:
        if not allow_none:
            raise TypeError(f"Argument {name} cannot be None.")
        return

    if not isinstance(value, expected_types):
        raise TypeError(
            f"Argument {name} '{str(value)}' is not supported. "
            f"Only values of type {str(expected_types)} are allowed."
        )


def validate_iterable(
    name,
    value,
    expected_element_types,
    allowed_iterables=(list, set),
    allow_none=False,
):
    """
    Validate that a value is an iterable (e.g., list, set) with elements of
    the specified type(s).

    Args:
        name (str): Name of the argument for error messages.
        value: Value to validate.
        expected_element_types (type or tuple of types): Expected type(s)
            for the elements.
        allowed_iterables (tuple of types): Iterable types that are
            allowed (default: list and set).
        allow_none (bool): Whether to allow the value to be None.
            Defaults to False.

    Raises:
        TypeError: If the value is not an allowed iterable type or if
            any element is not of the expected type(s).
    """

    if value is None:
        if not allow_none:
            raise TypeError(f"Argument {name} cannot be None.")
        return

    if not isinstance(value, allowed_iterables):
        raise TypeError(
            f"Argument {name} '{str(value)}' is not supported. "
            f"Only values of type {str(allowed_iterables)} are allowed."
        )
    if not all(
        isinstance(element, expected_element_types) for element in value
    ):
        raise TypeError(
            f"Invalid elements in {name} '{str(value)}'. "
            f"Only elements of type {str(expected_element_types)} are allowed."
        )


def validate_comparator_pytorch(name, value):
    """
    Validate that a value is a callable PyTorch comparator function.

    Args:
        name (str): Name of the argument for error messages.
        value: Value to validate.

    Raises:
        TypeError: If the value is not callable or not a PyTorch comparator.
    """

    # List of valid PyTorch comparator functions
    pytorch_comparators = {torch.gt, torch.lt, torch.ge, torch.le}

    # Check if value is callable and if it's one of
    # the PyTorch comparator functions
    if not callable(value):
        raise TypeError(
            f"Argument {name} '{str(value)}' is not supported. "
            "Only callable functions are allowed."
        )

    if value not in pytorch_comparators:
        raise TypeError(
            f"Argument {name} '{str(value)}' is not a valid PyTorch comparator "
            "function. Only PyTorch functions like torch.gt, torch.lt, "
            "torch.ge, torch.le are allowed."
        )


def validate_callable(name, value, allow_none=False):
    """
    Validate that a value is callable function.

    Args:
        name (str): Name of the argument for error messages.
        value: Value to validate.
        allow_none (bool): Whether to allow the value to be None.
            Defaults to False.

    Raises:
        TypeError: If the value is not callable.
    """

    if value is None:
        if not allow_none:
            raise TypeError(f"Argument {name} cannot be None.")
        return

    if not callable(value):
        raise TypeError(
            f"Argument {name} '{str(value)}' is not supported. "
            "Only callable functions are allowed."
        )


def validate_loaders() -> None:
    """
    TODO: implement function
    """
    # TODO complete
    pass



import random

import torch
from torch import ge, le
from torch.nn import MSELoss
from torch.optim import Adam
from torch.utils.tensorboard import SummaryWriter

from congrads.checkpoints import CheckpointManager
from congrads.constraints import Constraint, ScalarConstraint, SumConstraint
from congrads.core import CongradsCore
from congrads.datasets import FamilyIncome
from congrads.descriptor import Descriptor
from congrads.metrics import MetricManager
from congrads.networks import MLPNetwork
from congrads.transformations import DenormalizeMinMax
from congrads.utils import (
    CSVLogger,
    preprocess_FamilyIncome,
    split_data_loaders,
)

if __name__ == "__main__":

    # Set seed for reproducibility
    random.seed(42)
    seeds = []
    for i in range(3):
        seeds.append(random.randint(10, 10**6))
    torch.manual_seed(seeds[0])
    torch.cuda.manual_seed(seeds[1])
    torch.cuda.manual_seed_all(seeds[2])

    # CUDA for PyTorch
    use_cuda = torch.cuda.is_available()
    device = torch.device("cuda:0" if use_cuda else "cpu")
    torch.backends.cudnn.benchmark = True

    # Load and preprocess data
    data = FamilyIncome("./datasets", preprocess_FamilyIncome, download=True)
    loaders = split_data_loaders(
        data,
        loader_args={
            "batch_size": 100,
            "shuffle": True,
            "num_workers": 6,
            "prefetch_factor": 2,
        },
        valid_loader_args={"shuffle": False},
        test_loader_args={"shuffle": False},
    )

    # Instantiate network and push to correct device
    network = MLPNetwork(
        n_inputs=24, n_outputs=8, n_hidden_layers=6, hidden_dim=80
    )
    network = network.to(device)

    # Instantiate loss and optimizer
    criterion = MSELoss()
    optimizer = Adam(network.parameters(), lr=0.0001)

    # Descriptor setup
    descriptor = Descriptor()
    descriptor.add("input", 0, "Total Household Income", constant=True)
    descriptor.add("output", 0, "Total Food Expenditure")
    descriptor.add("output", 1, "Bread and Cereals Expenditure")
    descriptor.add("output", 2, "Meat Expenditure")
    descriptor.add("output", 3, "Vegetables Expenditure")
    descriptor.add("output", 4, "Housing and water Expenditure")
    descriptor.add("output", 5, "Medical Care Expenditure")
    descriptor.add("output", 6, "Communication Expenditure")
    descriptor.add("output", 7, "Education Expenditure")

    # Constraints definition
    Constraint.descriptor = descriptor
    Constraint.device = device
    constraints = [
        ScalarConstraint(
            "Total Food Expenditure",
            ge,
            0,
        ),
        ScalarConstraint("Total Food Expenditure", le, 1),
        ScalarConstraint("Bread and Cereals Expenditure", ge, 0),
        ScalarConstraint("Bread and Cereals Expenditure", le, 1),
        ScalarConstraint("Meat Expenditure", ge, 0),
        ScalarConstraint("Meat Expenditure", le, 1),
        ScalarConstraint("Vegetables Expenditure", ge, 0),
        ScalarConstraint("Vegetables Expenditure", le, 1),
        ScalarConstraint("Housing and water Expenditure", ge, 0),
        ScalarConstraint("Housing and water Expenditure", le, 1),
        ScalarConstraint("Medical Care Expenditure", ge, 0),
        ScalarConstraint("Medical Care Expenditure", le, 1),
        ScalarConstraint("Communication Expenditure", ge, 0),
        ScalarConstraint("Communication Expenditure", le, 1),
        ScalarConstraint("Education Expenditure", ge, 0),
        ScalarConstraint("Education Expenditure", le, 1),
        SumConstraint(
            [
                DenormalizeMinMax(
                    "Total Food Expenditure", min=3704, max=791848
                ),
            ],
            ge,
            [
                DenormalizeMinMax(
                    "Bread and Cereals Expenditure", min=0, max=437467
                ),
                DenormalizeMinMax("Meat Expenditure", min=0, max=140992),
                DenormalizeMinMax("Vegetables Expenditure", min=0, max=74800),
            ],
        ),
        SumConstraint(
            [
                DenormalizeMinMax(
                    "Total Household Income", min=11285, max=11815988
                ),
            ],
            ge,
            [
                DenormalizeMinMax(
                    "Total Food Expenditure", min=3704, max=791848
                ),
                DenormalizeMinMax(
                    "Housing and water Expenditure", min=1950, max=2188560
                ),
                DenormalizeMinMax(
                    "Medical Care Expenditure", min=0, max=1049275
                ),
                DenormalizeMinMax(
                    "Communication Expenditure", min=0, max=149940
                ),
                DenormalizeMinMax("Education Expenditure", min=0, max=731000),
            ],
        ),
    ]

    # Initialize metric manager
    metric_manager = MetricManager()

    # Initialize checkpoint manager
    checkpoint_manager = CheckpointManager(
        network,
        optimizer,
        metric_manager,
        save_dir="checkpoints/FamilyIncome",
        create_dir=True,
    )
    checkpoint_manager.register("CSR/valid")

    # Instantiate core
    core = CongradsCore(
        descriptor,
        constraints,
        loaders,
        network,
        criterion,
        optimizer,
        metric_manager,
        device,
        checkpoint_manager,
    )

    # Initialize data loggers
    tensorboard_logger = SummaryWriter(log_dir="logs/FamilyIncome")
    csv_logger = CSVLogger("logs/FamilyIncome.csv")

    def on_epoch_end(epoch: int):
        # Log metric values to TensorBoard and CSV file
        for name, value in metric_manager.aggregate("during_training").items():
            tensorboard_logger.add_scalar(name, value.item(), epoch)
            csv_logger.add_value(name, value.item(), epoch)

        # Write changes to disk
        tensorboard_logger.flush()
        csv_logger.save()

        # Reset metric manager
        metric_manager.reset("during_training")

        # Halve learning rate each 5 epochs
        if epoch % 5:
            for g in optimizer.param_groups:
                g["lr"] /= 2

    def on_train_end(epoch: int):
        # Log metric values to TensorBoard and CSV file
        for name, value in metric_manager.aggregate("after_training").items():
            tensorboard_logger.add_scalar(name, value.item(), epoch)
            csv_logger.add_value(name, value.item(), epoch)

        # Write changes to disk
        tensorboard_logger.flush()
        csv_logger.save()

        # Reset metric manager
        metric_manager.reset("after_training")

    # Start training
    start_epoch = checkpoint_manager.resume(ignore_missing=True)
    core.fit(
        start_epoch=start_epoch,
        max_epochs=50,
        on_epoch_end=on_epoch_end,
        on_train_end=on_train_end,
    )

    # Close writer
    tensorboard_logger.close()



import random

import torch
from torch import ge, gt, le
from torch.nn import MSELoss
from torch.optim import Adam
from torch.utils.tensorboard import SummaryWriter

from congrads.checkpoints import CheckpointManager
from congrads.constraints import BinaryConstraint, Constraint, ScalarConstraint
from congrads.core import CongradsCore
from congrads.datasets import BiasCorrection
from congrads.descriptor import Descriptor
from congrads.metrics import MetricManager
from congrads.networks import MLPNetwork
from congrads.utils import (
    CSVLogger,
    preprocess_BiasCorrection,
    split_data_loaders,
)


def test_BiasCorrection():

    # Set seed for reproducability
    random.seed(42)
    seeds = []
    for i in range(3):
        seeds.append(random.randint(10, 10**6))
    torch.manual_seed(seeds[0])
    torch.cuda.manual_seed(seeds[1])
    torch.cuda.manual_seed_all(seeds[2])

    # CUDA for PyTorch
    use_cuda = torch.cuda.is_available()
    device = torch.device("cuda:0" if use_cuda else "cpu")
    torch.backends.cudnn.benchmark = True

    # Load and preprocess data
    data = BiasCorrection(
        "./datasets", preprocess_BiasCorrection, download=True
    )
    loaders = split_data_loaders(
        data,
        loader_args={
            "batch_size": 100,
            "shuffle": True,
            "num_workers": 6,
            "prefetch_factor": 2,
        },
        valid_loader_args={"shuffle": False},
        test_loader_args={"shuffle": False},
    )

    # Instantiate network and push to correct device
    network = MLPNetwork(25, 2, n_hidden_layers=3, hidden_dim=35)
    network = network.to(device)

    # Instantiate loss and optimizer
    criterion = MSELoss()
    optimizer = Adam(network.parameters(), lr=0.001)

    # Descriptor setup
    descriptor = Descriptor()
    descriptor.add("output", 0, "Tmax")
    descriptor.add("output", 1, "Tmin")

    # Constraints definition
    Constraint.descriptor = descriptor
    Constraint.device = device
    constraints = [
        ScalarConstraint("Tmin", ge, 0),
        ScalarConstraint("Tmin", le, 1),
        ScalarConstraint("Tmax", ge, 0),
        ScalarConstraint("Tmax", le, 1),
        BinaryConstraint("Tmax", gt, "Tmin"),
    ]

    # Initialize metric manager
    metric_manager = MetricManager()

    # Initialize checkpoint manager
    checkpoint_manager = CheckpointManager(
        network,
        optimizer,
        metric_manager,
        save_dir="checkpoints/BiasCorrection",
        create_dir=True,
    )
    checkpoint_manager.register("CSR/valid")

    # Instantiate core
    core = CongradsCore(
        descriptor,
        constraints,
        loaders,
        network,
        criterion,
        optimizer,
        metric_manager,
        device,
        checkpoint_manager,
    )

    # Initialize data loggers
    tensorboard_logger = SummaryWriter(log_dir="logs/BiasCorrection")
    csv_logger = CSVLogger("logs/BiasCorrection.csv")

    def on_epoch_end(epoch: int):
        # Log metric values to TensorBoard and CSV file
        for name, value in metric_manager.aggregate("during_training").items():
            tensorboard_logger.add_scalar(name, value.item(), epoch)
            csv_logger.add_value(name, value.item(), epoch)

        # Write changes to disk
        tensorboard_logger.flush()
        csv_logger.save()

        # Reset metric manager
        metric_manager.reset("during_training")

        # Halve learning rate each 5 epochs
        if epoch % 5:
            for g in optimizer.param_groups:
                g["lr"] /= 2

    def on_train_end(epoch: int):
        # Log metric values to TensorBoard and CSV file
        for name, value in metric_manager.aggregate("after_training").items():
            tensorboard_logger.add_scalar(name, value.item(), epoch)
            csv_logger.add_value(name, value.item(), epoch)

        # Write changes to disk
        tensorboard_logger.flush()
        csv_logger.save()

        # Reset metric manager
        metric_manager.reset("after_training")

    # Start/resume training
    start_epoch = checkpoint_manager.resume(ignore_missing=True)
    core.fit(
        start_epoch=start_epoch,
        max_epochs=5,
        on_epoch_end=on_epoch_end,
        on_train_end=on_train_end,
    )

    # Close writer
    tensorboard_logger.close()
